@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2020 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -17,19 +15,20 @@
 
 """This module provides the core primitives of Hypothesis, such as given."""
 
-
-from __future__ import absolute_import, division, print_function
-
 import base64
 import contextlib
 import datetime
 import inspect
+import io
 import random as rnd_module
 import sys
 import traceback
 import warnings
 import zlib
+from inspect import getfullargspec
+from io import StringIO
 from random import Random
+from typing import Any, BinaryIO, Callable, Hashable, List, Optional, TypeVar, Union
 from unittest import TestCase
 
 import attr
@@ -59,13 +58,9 @@ from hypothesis.errors import (
 )
 from hypothesis.executors import new_style_executor
 from hypothesis.internal.compat import (
-    PY2,
     bad_django_TestCase,
     benchmark_time,
-    binary_type,
     get_type_hints,
-    getfullargspec,
-    hbytes,
     int_from_bytes,
     qualname,
 )
@@ -81,6 +76,7 @@ from hypothesis.internal.reflection import (
     arg_string,
     convert_positional_arguments,
     define_function_signature,
+    deprecated_posargs,
     function_digest,
     get_pretty_function_description,
     impersonate,
@@ -93,21 +89,18 @@ from hypothesis.reporting import (
     verbose_report,
     with_reporter,
 )
-from hypothesis.statistics import note_engine_for_statistics
+from hypothesis.statistics import describe_targets, note_engine_for_statistics
 from hypothesis.strategies._internal.collections import TupleStrategy
 from hypothesis.strategies._internal.strategies import (
+    Ex,
     MappedSearchStrategy,
     SearchStrategy,
 )
-from hypothesis.utils.conventions import infer
-from hypothesis.vendor.pretty import CUnicodeIO, RepresentationPrinter
+from hypothesis.utils.conventions import InferType, infer
+from hypothesis.vendor.pretty import RepresentationPrinter
 from hypothesis.version import __version__
 
-if False:
-    from typing import Any, Dict, Callable, Hashable, Optional, Union, TypeVar  # noqa
-    from hypothesis.utils.conventions import InferType  # noqa
-
-    TestFunc = TypeVar("TestFunc", bound=Callable)
+TestFunc = TypeVar("TestFunc", bound=Callable)
 
 
 running_under_pytest = False
@@ -115,13 +108,12 @@ global_force_seed = None
 
 
 @attr.s()
-class Example(object):
+class Example:
     args = attr.ib()
     kwargs = attr.ib()
 
 
-def example(*args, **kwargs):
-    # type: (*Any, **Any) -> Callable[[TestFunc], TestFunc]
+def example(*args: Any, **kwargs: Any) -> Callable[[TestFunc], TestFunc]:
     """A decorator which ensures a specific example is always tested."""
     if args and kwargs:
         raise InvalidArgument(
@@ -139,8 +131,7 @@ def example(*args, **kwargs):
     return accept
 
 
-def seed(seed):
-    # type: (Hashable) -> Callable[[TestFunc], TestFunc]
+def seed(seed: Hashable) -> Callable[[TestFunc], TestFunc]:
     """seed: Start the test execution from a specific seed.
 
     May be any hashable object. No exact meaning for seed is provided
@@ -187,9 +178,7 @@ def reproduce_failure(version, blob):
 
 
 def encode_failure(buffer):
-    # This needs to be a real bytes() instance, so we use binary_type()
-    # instead of hbytes() here.
-    buffer = binary_type(buffer)
+    buffer = bytes(buffer)
     compressed = zlib.compress(buffer)
     if len(compressed) < len(buffer):
         buffer = b"\1" + compressed
@@ -285,12 +274,6 @@ def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
             "%s() got an unexpected keyword argument %r, from `%s=%r` in @given"
             % (name, arg, arg, given_kwargs[arg])
         )
-    for a in original_argspec.args:
-        if isinstance(a, list):  # pragma: no cover
-            return invalid(
-                "Cannot decorate function %s() because it has destructuring arguments"
-                % (name,)
-            )
     if original_argspec.defaults or original_argspec.kwonlydefaults:
         return invalid("Cannot apply @given to a function with defaults.")
     missing = [repr(kw) for kw in original_argspec.kwonlyargs if kw not in given_kwargs]
@@ -313,13 +296,10 @@ class ArtificialDataForExample(ConjectureData):
     def __init__(self, kwargs):
         self.__draws = 0
         self.__kwargs = kwargs
-
-        super(ArtificialDataForExample, self).__init__(
-            max_length=0, prefix=hbytes(), parameter=None,
-        )
+        super().__init__(max_length=0, prefix=b"", random=None)
 
     def draw_bits(self, n):
-        raise NotImplementedError()  # pragma: no cover
+        raise NotImplementedError("Dummy object should never be asked for bits.")
 
     def draw(self, strategy):
         assert self.__draws == 0
@@ -367,6 +347,13 @@ def execute_explicit_examples(state, wrapped_test, arguments, kwargs):
                         is_final=True,
                         print_example=True,
                     )
+            except UnsatisfiedAssumption:
+                # Odd though it seems, we deliberately support explicit examples that
+                # are then rejected by a call to `assume()`.  As well as iterative
+                # development, this is rather useful to replay Hypothesis' part of
+                # a saved failure when other arguments are supplied by e.g. pytest.
+                # See https://github.com/HypothesisWorks/hypothesis/issues/2125
+                pass
             except BaseException:
                 report_buffered()
                 raise
@@ -397,7 +384,7 @@ def get_random_for_wrapped_test(test, wrapped_test):
 
 
 def process_arguments_to_given(
-    wrapped_test, arguments, kwargs, given_kwargs, argspec, test, settings,
+    wrapped_test, arguments, kwargs, given_kwargs, argspec, settings,
 ):
     selfy = None
     arguments, kwargs = convert_positional_arguments(wrapped_test, arguments, kwargs)
@@ -488,7 +475,7 @@ def new_given_argspec(original_argspec, given_kwargs):
     )
 
 
-class StateForActualGivenExecution(object):
+class StateForActualGivenExecution:
     def __init__(
         self, test_runner, search_strategy, test, settings, random, wrapped_test,
     ):
@@ -567,7 +554,7 @@ class StateForActualGivenExecution(object):
                             text_repr[0] = arg_string(test, args, kwargs)
 
                         if print_example or current_verbosity() >= Verbosity.verbose:
-                            output = CUnicodeIO()
+                            output = StringIO()
 
                             printer = RepresentationPrinter(output)
                             if print_example:
@@ -773,6 +760,11 @@ class StateForActualGivenExecution(object):
 
         flaky = 0
 
+        if runner.best_observed_targets:
+            for line in describe_targets(runner.best_observed_targets):
+                report(line)
+            report("")
+
         for falsifying_example in self.falsifying_examples:
             info = falsifying_example.extra_information
 
@@ -879,7 +871,7 @@ def fake_subTest(self, msg=None, **__):
 
 
 @attr.s()
-class HypothesisHandle(object):
+class HypothesisHandle:
     """This object is provided as the .hypothesis attribute on @given tests.
 
     Downstream users can reassign its attributes to insert custom logic into
@@ -895,13 +887,13 @@ class HypothesisHandle(object):
     """
 
     inner_test = attr.ib()
+    fuzz_one_input = attr.ib()
 
 
 def given(
-    *_given_arguments,  # type: Union[SearchStrategy, InferType]
-    **_given_kwargs  # type: Union[SearchStrategy, InferType]
-):
-    # type: (...) -> Callable[[Callable[..., None]], Callable[..., None]]
+    *_given_arguments: Union[SearchStrategy, InferType],
+    **_given_kwargs: Union[SearchStrategy, InferType]
+) -> Callable[[Callable[..., None]], Callable[..., None]]:
     """A decorator for turning a test function that accepts arguments into a
     randomized test.
 
@@ -940,6 +932,27 @@ def given(
 
         argspec = new_given_argspec(original_argspec, given_kwargs)
 
+        # Use type information to convert "infer" arguments into appropriate strategies.
+        if infer in given_kwargs.values():
+            hints = get_type_hints(test)
+        for name in [name for name, value in given_kwargs.items() if value is infer]:
+            if name not in hints:
+                # As usual, we want to emit this error when the test is executed,
+                # not when it's decorated.
+
+                @impersonate(test)
+                @define_function_signature(test.__name__, test.__doc__, argspec)
+                def wrapped_test(*arguments, **kwargs):
+                    __tracebackhide__ = True
+                    raise InvalidArgument(
+                        "passed %s=infer for %s, but %s has no type annotation"
+                        % (name, test.__name__, name)
+                    )
+
+                return wrapped_test
+
+            given_kwargs[name] = st.from_type(hints[name])
+
         @impersonate(test)
         @define_function_signature(test.__name__, test.__doc__, argspec)
         def wrapped_test(*arguments, **kwargs):
@@ -965,22 +978,8 @@ def given(
 
             random = get_random_for_wrapped_test(test, wrapped_test)
 
-            # Use type information to convert "infer" arguments into appropriate
-            # strategies.
-            if infer in given_kwargs.values():
-                hints = get_type_hints(test)
-            for name in [
-                name for name, value in given_kwargs.items() if value is infer
-            ]:
-                if name not in hints:
-                    raise InvalidArgument(
-                        "passed %s=infer for %s, but %s has no type annotation"
-                        % (name, test.__name__, name)
-                    )
-                given_kwargs[name] = st.from_type(hints[name])
-
             processed_args = process_arguments_to_given(
-                wrapped_test, arguments, kwargs, given_kwargs, argspec, test, settings,
+                wrapped_test, arguments, kwargs, given_kwargs, argspec, settings,
             )
             arguments, kwargs, test_runner, search_strategy = processed_args
 
@@ -1055,9 +1054,6 @@ def given(
             # The next step is to use the Conjecture engine to run the test on
             # many different inputs.
 
-            if settings.max_examples <= 0:
-                return
-
             if not (
                 Phase.reuse in settings.phases or Phase.generate in settings.phases
             ):
@@ -1096,18 +1092,8 @@ def given(
                     # full of Hypothesis internals they don't care about.
                     # We have to do this inline, to avoid adding another
                     # internal stack frame just when we've removed the rest.
-                    if PY2:
-                        # Python 2 doesn't have Exception.with_traceback(...);
-                        # instead it has a three-argument form of the `raise`
-                        # statement.  Unfortunately this is a SyntaxError on
-                        # Python 3, and before Python 2.7.9 it was *also* a
-                        # SyntaxError to use it in a nested function so we
-                        # can't `exec` or `eval` our way out (BPO-21591).
-                        # So unless we break some versions of Python 2, none
-                        # of them get traceback elision.
-                        raise
-                    # On Python 3, we swap out the real traceback for our
-                    # trimmed version.  Using a variable ensures that the line
+                    #
+                    # Using a variable for our trimmed error ensures that the line
                     # which will actually appear in tracebacks is as clear as
                     # possible - "raise the_error_hypothesis_found".
                     the_error_hypothesis_found = e.with_traceback(
@@ -1115,6 +1101,45 @@ def given(
                     )
                     raise the_error_hypothesis_found
 
+        def fuzz_one_input(
+            buffer: Union[bytes, bytearray, memoryview, BinaryIO]
+        ) -> Optional[bytes]:
+            """Run the test as a fuzz target, driven with the `buffer` of bytes.
+
+            Returns None if buffer invalid for the strategy, canonical pruned
+            bytes if the buffer was valid, and leaves raised exceptions alone.
+
+            Note: this feature is experimental and may change or be removed.
+            """
+            if isinstance(buffer, io.IOBase):
+                buffer = buffer.read()
+            if isinstance(buffer, (bytearray, memoryview)):
+                buffer = bytes(buffer)
+            assert isinstance(buffer, bytes)
+
+            test = wrapped_test.hypothesis.inner_test
+            settings = wrapped_test._hypothesis_internal_use_settings
+            random = get_random_for_wrapped_test(test, wrapped_test)
+            _args, _kwargs, test_runner, search_strategy = process_arguments_to_given(
+                wrapped_test, (), {}, given_kwargs, argspec, settings,
+            )
+            assert not _args
+            assert not _kwargs
+            state = StateForActualGivenExecution(
+                test_runner, search_strategy, test, settings, random, wrapped_test,
+            )
+
+            data = ConjectureData.for_buffer(buffer)
+            try:
+                state.execute_once(data)
+            except (StopTest, UnsatisfiedAssumption):
+                return None
+            except BaseException:
+                if settings.database is not None:
+                    settings.database.save(function_digest(test), bytes(data.buffer))
+                raise
+            return bytes(data.buffer)
+
         # After having created the decorated test function, we need to copy
         # over some attributes to make the switch as seamless as possible.
 
@@ -1134,20 +1159,21 @@ def given(
         wrapped_test._hypothesis_internal_use_reproduce_failure = getattr(
             test, "_hypothesis_internal_use_reproduce_failure", None
         )
-        wrapped_test.hypothesis = HypothesisHandle(test)
+        wrapped_test.hypothesis = HypothesisHandle(test, fuzz_one_input)
         return wrapped_test
 
     return run_test_as_given
 
 
+@deprecated_posargs
 def find(
-    specifier,  # type: SearchStrategy
-    condition,  # type: Callable[[Any], bool]
-    settings=None,  # type: Settings
-    random=None,  # type: Any
-    database_key=None,  # type: bytes
-):
-    # type: (...) -> Any
+    specifier: SearchStrategy[Ex],
+    condition: Callable[[Any], bool],
+    *,
+    settings: Settings = None,
+    random: Random = None,
+    database_key: bytes = None
+) -> Ex:
     """Returns the minimal example from the given strategy ``specifier`` that
     matches the predicate function ``condition``."""
     if settings is None:
@@ -1166,13 +1192,13 @@ def find(
         )
     specifier.validate()
 
-    last = [None]
+    last = []  # type: List[Ex]
 
     @settings
     @given(specifier)
     def test(v):
         if condition(v):
-            last[0] = v
+            last[:] = [v]
             raise Found()
 
     if random is not None:
